from littledarwin.JavaParse import JavaParse
from antlr4 import *
from chaosmeter.metrics import Metric


class CyclomaticComplexity(Metric.Metric):
    instantiable = True
    name = "Cyclomatic Complexity"
    abbreviation = "CC"

    def __init__(self, javaParseInstance: JavaParse):
        super().__init__(javaParseInstance)
        self.defaultValue = 1

    def calculate(self, tree: RuleContext, sourceCode: str = ""):
        return self.javaParseInstance.getCyclomaticComplexityAllMethods(tree)



import os
from typing import Dict, List
from chaosmeter.writers.Writer import Writer


class CSVWriter(Writer):
    instantiable = True
    name = "CSV Writer"

    def __init__(self, outputPath):
        super().__init__(outputPath)
        self.extension = ".csv"

    def createTargetFormat(self, metricValues: Dict[str, List[int]], metricLabels: List[str]) -> bytes:
        fileContentLines = [';'.join(metricLabels)]

        for methodName in metricValues.keys():
            fileContentLines.append(methodName + ';' + ';'.join([str(x) for x in metricValues[methodName]]))

        fileContent = os.linesep.join(fileContentLines)
        return fileContent.encode()

    def createFinalReportTargetFormat(self, finalReport: List[list]) -> bytes:
        fileContentLines = [';'.join([str(x) for x in line]) for line in finalReport]
        fileContent = os.linesep.join(fileContentLines)
        return fileContent.encode()






import os

__all__ = list()
for fileName in os.listdir(os.path.dirname(__file__)):
    if fileName.endswith(".py") \
            and os.path.isfile(os.path.join(os.path.dirname(__file__), fileName)) \
            and not fileName.endswith('__init__.py'):

        __all__.append(os.path.basename(fileName[:-3]))




from littledarwin.JavaParse import JavaParse
from antlr4 import *
from chaosmeter.metrics import Metric


class SourceLinesOfCode(Metric.Metric):
    instantiable = True
    name = "Source Lines of Code"
    abbreviation = "SLOC"

    def __init__(self, javaParseInstance: JavaParse):
        super().__init__(javaParseInstance)

    def calculate(self, tree: RuleContext, sourceCode: str = ""):
        return self.javaParseInstance.getLinesOfCodePerMethod(tree)




from littledarwin.JavaParse import JavaParse
from littledarwin.JavaMutate import JavaMutate
from antlr4 import *
from chaosmeter.metrics import Metric


class MutantDensity(Metric.Metric):
    instantiable = True
    name = "Mutant Density"
    abbreviation = "MD"

    def __init__(self, javaParseInstance: JavaParse):
        super().__init__(javaParseInstance)

    def calculate(self, tree: RuleContext, sourceCode: str = ""):
        javaMutate = JavaMutate(tree, sourceCode, self.javaParseInstance, False)
        javaMutate.gatherMutants(["Traditional"])
        result = javaMutate.mutantsPerMethod.copy()

        del javaMutate

        return result






"""
__init__ script for chaosmeter package
"""

from chaosmeter.ChaosMeter import chaosMeterVersion

__version__ = chaosMeterVersion
__author__ = "Ali Parsai"
__package__ = 'chaosmeter'
__license__ = 'GNU GPLv3'
__url__ = 'https://chaosmeter.parsai.net/'





from io import BytesIO
from typing import Dict, List
from chaosmeter.writers.Writer import Writer

from odf.opendocument import OpenDocumentSpreadsheet
from odf.style import Style, TextProperties, ParagraphProperties
from odf.text import P
from odf.table import Table, TableRow, TableCell


class ODSWriter(Writer):
    instantiable = True
    name = "ODS Writer"

    def __init__(self, outputPath):
        super().__init__(outputPath)
        self.extension = ".ods"

    def createTargetFormat(self, metricValues: Dict[str, List[int]], metricLabels: List[str]) -> bytes:
        textdoc = OpenDocumentSpreadsheet()
        tablecontents = Style(name="Table Contents", family="paragraph")
        tablecontents.addElement(ParagraphProperties(numberlines="false", linenumber="0"))
        tablecontents.addElement(TextProperties(fontweight="bold"))
        textdoc.styles.addElement(tablecontents)

        table = Table(name="Java Metrics")

        tr = self.newRow(table)
        for metricLabel in metricLabels:
            self.addElementToRow(metricLabel, "string", tr, tablecontents)

        for methodName in metricValues.keys():
            tr = self.newRow(table)
            self.addElementToRow(methodName, "string", tr, tablecontents)

            for metricValue in metricValues[methodName]:
                self.addElementToRow(str(metricValue), "float", tr, tablecontents)

        textdoc.spreadsheet.addElement(table)

        stringOutput = BytesIO()
        textdoc.write(stringOutput)
        return stringOutput.getvalue()

    def createFinalReportTargetFormat(self, finalReport: List[list]) -> bytes:
        textdoc = OpenDocumentSpreadsheet()
        tablecontents = Style(name="Table Contents", family="paragraph")
        tablecontents.addElement(ParagraphProperties(numberlines="false", linenumber="0"))
        tablecontents.addElement(TextProperties(fontweight="bold"))
        textdoc.styles.addElement(tablecontents)

        table = Table(name="Java Metrics")

        tr = self.newRow(table)
        for columnLabels in finalReport[0]:
            self.addElementToRow(columnLabels, "string", tr, tablecontents)

        for row in finalReport[1:]:
            tr = self.newRow(table)
            self.addElementToRow(row[0], "string", tr, tablecontents)
            self.addElementToRow(row[1], "string", tr, tablecontents)

            for element in row[2:]:
                self.addElementToRow(str(element), "float", tr, tablecontents)

        textdoc.spreadsheet.addElement(table)

        stringOutput = BytesIO()
        textdoc.write(stringOutput)
        return stringOutput.getvalue()

    def newRow(self, table):
        tr = TableRow()
        table.addElement(tr)
        return tr

    def addElementToRow(self, element, elementType, tableRow, tablecontents):
        if elementType == "float":
            tc = TableCell(valuetype=elementType, value=element.strip())
        else:
            tc = TableCell(valuetype=elementType)

        tableRow.addElement(tc)
        p = P(stylename=tablecontents, text=element)
        tc.addElement(p)



################################################################################################################
##                                                                                                            ##
##                                                                                                            ##
##        ▄█▄     ▄  █ ██   ████▄    ▄▄▄▄▄   █▀▄▀█ ▄███▄      ▄▄▄▄▀ ▄███▄   █▄▄▄▄                             ##
##        █▀ ▀▄  █   █ █ █  █   █   █     ▀▄ █ █ █ █▀   ▀  ▀▀▀ █    █▀   ▀  █  ▄▀                             ##
##        █   ▀  ██▀▀█ █▄▄█ █   █ ▄  ▀▀▀▀▄   █ ▄ █ ██▄▄        █    ██▄▄    █▀▀▌                              ##
##        █▄  ▄▀ █   █ █  █ ▀████  ▀▄▄▄▄▀    █   █ █▄   ▄▀    █     █▄   ▄▀ █  █                              ##
##         ▀███▀     █     █                     █  ▀███▀     ▀      ▀███▀     █                              ##
##                  ▀     █                     ▀                             ▀                               ##
##                       ▀                                                                                    ##
##                                                                                                            ##
##       Copyright (c) 2020 Ali Parsai                                                                        ##
##                                                                                                            ##
##       This program is free software: you can redistribute it and/or modify it under the terms of           ##
##       the GNU General Public License as published by the Free Software Foundation, either version 3        ##
##       of the License, or (at your option) any later version.                                               ##
##                                                                                                            ##
##       This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;            ##
##       without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            ##
##       See the GNU General Public License for more details.                                                 ##
##                                                                                                            ##
##       You should have received a copy of the GNU General Public License along with this program.           ##
##       If not, see <https://www.gnu.org/licenses/>.                                                         ##
##                                                                                                            ##
##       Find me at:                                                                                          ##
##       https://www.parsai.net                                                                               ##
##                                                                                                            ##
################################################################################################################
import fnmatch
import os
import sys
from optparse import OptionParser, Values

from littledarwin import JavaParse
from littledarwin import JavaIO
from tqdm import tqdm

from .metrics import *
from .writers import *
from chaosmeter import License

chaosMeterVersion = '0.1.2'


def main(mockArgs: list = None):
    """
    Main ChaosMeter Function
    """
    print("""

▄█▄     ▄  █ ██   ████▄    ▄▄▄▄▄   █▀▄▀█ ▄███▄      ▄▄▄▄▀ ▄███▄   █▄▄▄▄
█▀ ▀▄  █   █ █ █  █   █   █     ▀▄ █ █ █ █▀   ▀  ▀▀▀ █    █▀   ▀  █  ▄▀
█   ▀  ██▀▀█ █▄▄█ █   █ ▄  ▀▀▀▀▄   █ ▄ █ ██▄▄        █    ██▄▄    █▀▀▌
█▄  ▄▀ █   █ █  █ ▀████  ▀▄▄▄▄▀    █   █ █▄   ▄▀    █     █▄   ▄▀ █  █
▀███▀     █     █                     █  ▀███▀     ▀      ▀███▀     █
         ▀     █                     ▀                             ▀
              ▀

    ChaosMeter version %s Copyright (C) 2020 Ali Parsai

    ChaosMeter comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it
    under certain conditions; run ChaosMeter --license for details.


    """ % chaosMeterVersion)

    optionParser = OptionParser(prog="chaosmeter")
    options = parseCmdArgs(optionParser, mockArgs)

    if options.sourcePath is None:
        optionParser.print_help()
        print("\nYou need to specify at least the path to the source files.\n")
        print("\nExample:\n\t ChaosMeter -p ./src/main -t ./target \n\n")
        sys.exit(1)

    if not os.path.isdir(options.sourcePath):
        print("Source path must be a directory.")
        sys.exit(2)

    sourcePath = os.path.abspath(options.sourcePath)
    targetPath = os.path.abspath(options.targetPath)

    javaParseInstance = JavaParse.JavaParse()
    javaIOInstance = JavaIO.JavaIO()

    metricList = Metric.getAllMetrics()
    metricInstanceList = list()

    if len(metricList) == 0:
        print("No metrics found!")
        sys.exit(3)

    # Find all metrics

    for MetricClass in metricList:
        metricInstance = MetricClass(javaParseInstance)
        metricInstanceList.append(metricInstance)
        print("Found metric: \"" + MetricClass.name + "\"")

    print(os.linesep)

    # Find all writers
    writerList = Writer.getAllWriters()
    writerInstanceList = list()

    if len(writerList) == 0:
        print("No writers found!")
        sys.exit(4)

    for WriterClass in writerList:
        writerInstance = WriterClass(targetPath)
        writerInstanceList.append(writerInstance)
        print("Found writer: \"" + WriterClass.name + "\"")

    # Get the file list

    if not os.path.isdir(sourcePath):
        print("Source path must be a directory.")
        sys.exit(5)

    fileList = list()

    for root, dirnames, filenames in os.walk(sourcePath):
        for filePath in fnmatch.filter(filenames, "*.java"):
            fileList.append(os.path.join(root, filePath))

    fileCounter = 0
    fileCount = len(fileList)

    print(os.linesep)
    print("Source Path: ", sourcePath)
    print("Target Path: ", targetPath)
    print(os.linesep)

    # Main loop
    completeResults = dict()
    completeResultsPath = os.path.join(targetPath, "FinalReport")
    for srcFile in tqdm(fileList, dynamic_ncols=True, unit='files'):
        fileCounter += 1
        fileRelativePath = os.path.relpath(srcFile, sourcePath)
        try:
            tqdm.write("({:,}/{:,}) {}".format(fileCounter, fileCount, fileRelativePath), end="\n\n")
        except UnicodeError as e:
            tqdm.write(str(e) + os.linesep)
            tqdm.write("Non-unicode filename detected. Not showing in terminal.")

        try:
            # parsing the source file into a tree.
            sourceCode = javaIOInstance.getFileContent(srcFile)
            tree = javaParseInstance.parse(sourceCode)
        except Exception as e:
            tqdm.write(str(e) + os.linesep, file=sys.stderr)
            tqdm.write("Error in parsing Java code, skipping the file.")
            continue

        # Calculate metrics

        metricResults = dict()
        for metricInstance in metricInstanceList:
            metricResults[metricInstance.name] = metricInstance.calculate(tree, sourceCode)

        metricResultsAggregate, metricLabels = Metric.aggregateMetrics(**metricResults)

        # Prepare the result file
        completeResults[fileRelativePath] = metricResultsAggregate

        srcFileRoot, srcFileName = os.path.split(srcFile)
        targetDir = os.path.join(targetPath, os.path.relpath(srcFileRoot, sourcePath))
        if not os.path.exists(targetDir):
            os.makedirs(targetDir)

        targetFilePath = os.path.splitext(os.path.join(targetDir, srcFileName))[0]

        for writerInstance in writerInstanceList:
            fileContent = writerInstance.createTargetFormat(metricResultsAggregate, metricLabels)
            writerInstance.write(targetFilePath, fileContent)

    completeResultsLabels = ["File"]
    completeResultsLabels.extend(metricLabels)
    completeResultsAggregate = [completeResultsLabels]

    for filePath in sorted(completeResults.keys()):
        for methodName in sorted(completeResults[filePath].keys()):
            cellList = [filePath, methodName]
            cellList.extend(completeResults[filePath][methodName])
            completeResultsAggregate.append(cellList)

    for writerInstance in writerInstanceList:
        completeFileContent = writerInstance.createFinalReportTargetFormat(completeResultsAggregate)
        writerInstance.write(completeResultsPath, completeFileContent)

    print(os.linesep)

    return 0


def parseCmdArgs(optionParser: OptionParser, mockArgs: list = None) -> Values:
    """

    :param mockArgs:
    :type mockArgs:
    :param optionParser:
    :type optionParser:
    :return:
    :rtype:
    """
    # parsing input options
    optionParser.add_option("-p", "--path", action="store", dest="sourcePath",
                            default=None, help="Path to Java source files.")
    optionParser.add_option("-t", "--target", action="store", dest="targetPath",
                            default=os.path.dirname(os.path.realpath(__file__)),
                            help="Path to store results.")
    optionParser.add_option("--license", action="store_true", dest="isLicenseActive", default=False,
                            help="Output the license and exit.")

    if mockArgs is None:
        (options, args) = optionParser.parse_args()
    else:
        (options, args) = optionParser.parse_args(args=mockArgs)

    if options.isLicenseActive:
        License.outputLicense()
        sys.exit(0)

    return options



#!/usr/bin/env python3

"""
__main__ script for chaosmeter package
"""

from chaosmeter import ChaosMeter
import sys


def entryPoint():
    ChaosMeter.main()


if __name__ == "__main__":
    sys.exit(entryPoint())



from typing import Set, List, Dict, Tuple

from antlr4 import *
from littledarwin.JavaParse import JavaParse


def getAllInstantiableSubclasses(parentClass):
    """

    :param parentClass: the class that all its subclasses must be returned
    :type parentClass: Type[MutationOperator]
    :return: set of MutationOperator instantiable subclasses
    :rtype: set
    """
    allInstantiableSubclasses = set()

    for subClass in parentClass.__subclasses__():
        if subClass.instantiable:
            allInstantiableSubclasses.add(subClass)
        allInstantiableSubclasses.update(getAllInstantiableSubclasses(subClass))

    return allInstantiableSubclasses


class Metric(object):
    instantiable = False
    name = "Metric"
    abbreviation = "M"

    def __init__(self, javaParseInstance: JavaParse):
        self.javaParseInstance = javaParseInstance
        self.defaultValue = 0

    def calculate(self, tree: RuleContext, sourceCode: str = ""):
        pass


def getAllMetrics() -> Set[Metric]:
    return getAllInstantiableSubclasses(Metric)


def aggregateMetrics(**kargs: Dict[str, int]) -> Tuple[Dict[str, List[int]], List[str]]:
    aggregate = dict()
    labels = ["Method Name"]
    metricList = sorted(kargs.keys())
    labels.extend([str(metricName) for metricName in metricList])

    methodList = set()

    for mL in metricList:
        methodList.update(kargs[mL].keys())

    for method in methodList:
        aggregate[method] = list()
        for mL in metricList:
            aggregate[method].append(kargs[mL].get(method, 0))
    return aggregate, labels



import os
from typing import Set, List, Dict


def getAllInstantiableSubclasses(parentClass):
    """

    :param parentClass: the class that all its subclasses must be returned
    :type parentClass: Type[MutationOperator]
    :return: set of MutationOperator instantiable subclasses
    :rtype: set
    """
    allInstantiableSubclasses = set()

    for subClass in parentClass.__subclasses__():
        if subClass.instantiable:
            allInstantiableSubclasses.add(subClass)
        allInstantiableSubclasses.update(getAllInstantiableSubclasses(subClass))

    return allInstantiableSubclasses


class Writer(object):
    instantiable = False
    name = "Writer"

    def __init__(self, outputPath):
        self.outputPath = outputPath
        self.extension = ".dummy"
        self.makeOutputPath()

    def makeOutputPath(self):
        if not os.path.exists(self.outputPath):
            os.makedirs(self.outputPath)

    def createTargetFormat(self, metricValues: Dict[str, List[int]], metricLabels: List[str]) -> bytes:
        pass

    def createFinalReportTargetFormat(self, finalReport: List[list]) -> bytes:
        pass

    def write(self, filePath: str, fileContent: bytes):
        with open(filePath + self.extension, 'wb') as outputHandle:
            outputHandle.write(fileContent)


def getAllWriters() -> Set[Writer]:
    return getAllInstantiableSubclasses(Writer)
